/*  PCX2GRAY v1.0 by Magnus Svedin -96
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define uint unsigned int
#define uchar unsigned char

#define BITS 3  /* 3=8-gray picture, 2=4-gray picture  */
#define TYPE 0  /* 0=write .85s-file, 1=write asm-file */

void checkfile(uchar *header);
uchar readpalette(FILE *f, uchar *buffer);
void createtable(uchar *pal, uchar *table);
void readimage(FILE *f, uchar *buffer);
void convert(uchar *imagedata, uchar *table);
void writetext(uchar *imagedata, char *filename);
void writestring(uchar *imagedata, char *filename);

void main(int antarg, char *arg[])
{
  FILE *f;
  uchar pal[768], image[128*64];
  int a;
  uchar grayvalues[256];
  char outfilename[20];
  printf("Pcx2gray v1.0 by Magnus Svedin -96\n");
  if(antarg != 2)
  {
    printf("Usage: pcx2gray [filename]\n");
    exit(1);
  }
  if((f=fopen(arg[1], "rb")) == NULL)
  {
    printf("Unable to open file: %s\n", arg[1]);
    exit(1);
  }
  fread(pal, 1, 70, f);
  checkfile(pal);
  if(!readpalette(f, pal))
  {
    printf("Error reading palette\n");
    exit(1);
  }
  createtable(pal, grayvalues);
  readimage(f, image);
  fclose(f);
  convert(image, grayvalues); 

  for(a=0;arg[1][a]!='.' && arg[1][a]!='\0';a++)
    outfilename[a]=arg[1][a];
  outfilename[a++]='.';
  outfilename[a++]='a';
  outfilename[a++]='s';
  outfilename[a++]='m';
  outfilename[a]='\0';

  
  if(TYPE==0)
    writestring(image, outfilename);
  else
    writetext(image, outfilename); 

  printf("%s written ok.\n", outfilename);
}

void writestring(uchar *imagedata, char *filename)
{
  uchar curbit, curbyte;
  FILE *f;
  int imageptr, checksum=0x37d, a;
  char header[74]={'*', '*', 'T', 'I', '8', '5', '*', '*', 0x1a, 0x0c, 0x00,
                  '8', '-', 'G', 'r', 'a', 'y', ' ', 'p', 'i', 'c', ' ', 'c',
                  'o', 'n', 'v', 'e', 'r', 't', 'e', 'd', ' ', 'w', 'i', 't',
                  'h', ' ', 'p', 'i', 'c', '2', 'g', 'r', 'a', 'y', ' ', 'b',
                  'y', ' ', 'M', 'S', 'v', '.',  0x14, 0x0c, 0x0b, 0x00,
                  0x05, 0x0c, 0x0c, 0x07, 'G', 'r', 'a', 'y', 'P', 'i', 'c',
                  0x05, 0x0c, 0x03, 0x0c, 0x00, 0x7f};

  filename[strlen(filename)-1]='s';
  filename[strlen(filename)-2]='5';
  filename[strlen(filename)-3]='8';
  if((f=fopen(filename, "wb")) == NULL)
  {
    printf("Unable to write output-file\n");
    exit(1);
  }
  fwrite(header, 1, 74, f);
  for(curbit=(1 << (BITS-1));curbit != 0;curbit=(curbit >> 1))
  {
    imageptr=0;
    while(imageptr < 128*64)
    {
      curbyte=0;
      for(a=0;a<8;a++)
      {
        if((imagedata[imageptr++] & curbit) != 0)
          curbyte=curbyte | (0x80 >> a);
      }
      checksum+=(int)curbyte;
      fwrite(&curbyte, 1, 1, f);
    }
  }
  header[0]=0x00;  /* Not quite sure why this byte is here  */
  header[1]=(checksum & 0xff);
  header[2]=((checksum & 0xff00) >> 8);
  fwrite(header, 1, 3, f);
  fclose(f);
}

void writetext(uchar *imagedata, char *filename)
{
  uchar curbit, curbyte;
  FILE *f;
  int imageptr, a, bitnr=1;
  if((f=fopen(filename, "wt")) == NULL)
  {
    printf("Unable to write output-file\n");
    exit(1);
  }
  for(curbit=1<<(BITS-1);curbit != 0;curbit=(curbit >> 1))
  {
    imageptr=0;
    fprintf(f, "GrayPic_bit%d:\n\t.db ", bitnr++);
    while(imageptr < 128*64)
    {
      curbyte=0;
      for(a=0;a<8;a++)
      {
        if((imagedata[imageptr++] & curbit) != 0)
          curbyte=curbyte | (0x80 >> a);
      }
      fprintf(f, "0%xh", curbyte);
      if((((int)((double)imageptr/8)) % 16) == 0)
        if(imageptr < 128*64)
          fprintf(f, "\n\t.db ");
        else
          fprintf(f, "\n");
      else
        fprintf(f, ", ");
    }
  }
  fclose(f);
}

void convert(uchar *imagedata, uchar *table)
{
  int a;
  for(a=0;a<128*64;a++)
    imagedata[a]=table[imagedata[a]];
}

void readimage(FILE *f, uchar *buffer)
{
  uchar *pcxbuffer;
  int a=0, pixels=0, datalen, counter, data;
  fseek(f, -769, SEEK_END);
  datalen=ftell(f)-128;
  pcxbuffer=(uchar *)malloc(datalen);
  fseek(f, 128, SEEK_SET);
  fread(pcxbuffer, 1, datalen, f);
  while(pixels < 128*64)
  {
    if((pcxbuffer[a] & 0xc0) == 0xc0)
    {
      counter=pcxbuffer[a] & 0x3f;
      data=pcxbuffer[a+1];
      for(;counter>0;counter--)
        buffer[pixels++]=data;
      a=a+2;
    }
    else
      buffer[pixels++]=pcxbuffer[a++];
  }
}

void createtable(uchar *pal, uchar *table)
{
  int a;
  for(a=0;a<256;a++)
    table[a]=((0xff << BITS)^0Xff)-(((pal[a*3]+pal[a*3+1]+pal[a*3+2])/3)>>(8-BITS));
}

uchar readpalette(FILE *f, uchar *buffer)
{
  uchar check;
  fseek(f, -769, SEEK_END);
  fread(&check, 1, 1, f);
  if(check!=12)
    return(0);
  fread(buffer, 1, 768, f);
  return(1);
}

void checkfile(uchar *header)
{
  if(header[1]!=5)
  {
    printf("File must be pcx-version 3.0\n");
    exit(1);
  }
  if(header[2]!=1)
  {
    printf("File encoding must be PCX run-length encoding\n");
    exit(1);
  }
  if(header[3]!=8)
  {
    printf("File must be an 8-bit pcx-file\n");
    exit(1);
  }
  if(header[8]-header[4]!=127 || header[10]-header[6]!=63)
  {
    printf("Size of picture must be 128x64\n");
    exit(1);
  }
  
}
